import logging
from abc import ABC, abstractmethod
from base64 import b64encode
from typing import Self

import validators
from httpx import Client, Cookies, BasicAuth, Limits, AsyncClient, Request

from .enums import AuthFlow, HttpMethods
from .enums import JoinTypes, OrderByDirections
from .exceptions import (
    ApiAuthError,
    ApiError,
    ApiException,
    SslException,
)
from .exceptions import BuildQueryException

_ACTION_TO_METHOD_MAPPING = {
    "get": HttpMethods.GET,
    "create": HttpMethods.POST,
    "update": HttpMethods.POST,
}

logger = logging.getLogger(__package__)


def get_method(action: str) -> HttpMethods:
    """
    Return the method HttpMethod Enum for an action
    :param action:
    """
    return _ACTION_TO_METHOD_MAPPING.get(action)


# noinspection PyAttributeOutsideInit,PyUnresolvedReferences
class Api4:
    """
    Generic class for API v4 entities.
    """

    JOIN_TYPES = JoinTypes
    ORDER_BY_DIRECTIONS = OrderByDirections

    def __init__(self):
        """
        Do not use the constructor directly. Use an action as a factory method.
        """
        raise RuntimeError("Use a factory method like 'get()' to create an instance.")

    @classmethod
    def _create_instance(cls, entity: str, action: str):
        instance = object.__new__(cls)
        instance._entity = entity
        instance._action = action
        instance._check_permissions = True
        instance._limit = 0
        instance._offset = 0
        instance._language = None
        instance._translation_mode = None
        instance._selects = []
        instance._joins = []
        instance._wheres = []
        instance._group_by = []
        instance._order_by = {}
        instance._having = []
        instance._chains = {}
        return instance

    @classmethod
    def get(cls, entity: str, check_permissions: bool = True):
        """
        Initialize a new get query
        :param entity: A CiviCRM entity
        :param check_permissions:
        :return:
        """
        instance = cls._create_instance(entity, "get")
        instance._check_permissions = check_permissions
        return instance

    def set_limit(self, limit: int):
        """
        Set the limit for the query
        :param limit:
        :return:
        """
        self._limit = limit
        return self

    def set_offset(self, offset: int):
        """
        Set the offset for the query
        :param offset:
        :return:
        """
        self._offset = offset
        return self

    def set_language(self, language: str):
        """
        Set the language for the query
        :param language: e.g. 'de_DE'
        :return:
        """
        self._language = language
        return self

    def set_translation_mode(self, mode: str):
        """
        Set the translation mode for the query
        :param mode: 'fuzzy' or 'strict'
        :return:
        """
        self._translation_mode = mode
        return

    def select_row_count(self):
        """
        Add the row count to the select statement
        :return:
        """
        self._selects.append("row_count")
        return

    def add_select(self, key: str):
        """
        Add a key to the select statement
        :param key:
        :return:
        """
        self._selects.append(key)
        return self

    def set_select(self, select: list):
        """
        Set the full select statement
        :param select:
        :return:
        """
        self._selects = select
        return self

    def add_join(self, entity: str, join_type: JoinTypes, on_clause: list):
        """
        Add a join clause
        :param on_clause:
        :param join_type:
        :param entity:
        :return:
        """
        self._joins.append([entity, join_type.value, on_clause])
        return self

    def set_join(self, joins: list):
        """
        Set the full join statement
        :param joins:
        :return:
        """
        self._joins = joins
        return self

    def add_where(self, where: list):
        """
        Add a where clause
        :param where:
        :return:
        """
        self._wheres.append(where)
        return self

    def set_where(self, where: list):
        """
        Set the full where clause
        :param where:
        :return:
        """
        self._wheres = where
        return self

    def add_group_by(self, key: str):
        """
        Add a group by key
        :param key:
        :return:
        """
        self._group_by.append(key)
        return self

    def set_group_by(self, group_by: list):
        """
        Set the full group by statement
        :param group_by:
        :return:
        """
        self._group_by = group_by
        return self

    def add_having(self, having: list):
        """
        Add a having clause
        :param having:
        :return:
        """
        self._having.append(having)
        return self

    def set_having(self, having: list):
        """
        Set the full having clause
        :param having:
        :return:
        """
        self._having = having
        return self

    def add_order_by(self, key: str, direction: str):
        """
        Add an order-by key
        :param key:
        :param direction: 'ASC' or 'DESC'
        :return:
        """
        self._order_by.update({key: direction})
        return self

    def set_order_by(self, order_by: dict):
        """
        Set the full order by statement
        :param order_by:
        :return:
        """
        self._order_by = order_by
        return self

    def add_chain(self, name: str, entity: Self):
        """
        Add a chained request
        :param name: Name of the chained request
        :param entity: Api4 instance (do not execute it)
        :return:
        """
        self._chains.update({name: entity})
        return self

    def set_chain(self, chain: dict):
        """
        Set the full chain statement
        :param chain:
        :return:
        """
        self._chains = chain
        return self

    def execute(self) -> dict:
        """
        Execute the query
        :return: Response
        """
        from . import api

        method = get_method(self._action)
        return api.execute_api4(method, self)

    def aexecute(self) -> dict:
        """
        Execute the query asynchronously
        :return: Response
        """
        from . import async_api

        method = get_method(self._action)
        return async_api.execute_api4(method, self)

    @property
    def entity(self) -> str:
        """
        Get the entity
        :return: str
        """
        return self._entity

    @property
    def action(self) -> str:
        """
        Get the action
        :return: str
        """
        return self._action

    def build_query(self) -> dict:
        """
        Build the query
        :return: dict
        """

        if not self._entity:
            raise BuildQueryException("No entity provided")

        if not self._action:
            raise BuildQueryException("No action provided")

        # Build chain
        chain = {}
        for c in self._chains:
            chain_name = c.get("name")
            entity = c.get("entity")
            chain.update(
                entity.build_query(
                    {chain_name: [entity.entity, entity.action, entity.build_query()]}
                )
            )

        # Build query
        query = {"checkPermissions": self._check_permissions}
        if language := self._language:
            query["language"] = language
        if translation_mode := self._translation_mode:
            query["translationMode"] = translation_mode
        if limit := self._limit:
            query["limit"] = limit
        if offset := self._offset:
            query["offset"] = offset
        if select := self._selects or ["*"]:
            query["select"] = select
        if join := self._joins:
            query["join"] = join
        if where := self._wheres:
            query["where"] = where
        if group_by := self._group_by:
            query["groupBy"] = group_by
        if order_by := self._order_by:
            query["orderBy"] = order_by
        if having := self._having:
            query["having"] = having
        if chain:
            query["chain"] = chain

        return query


class BaseCiviCrmApi(ABC):
    """
    This is the base class for the CiviCRM API.
    """

    AUTH_FLOWS = AuthFlow
    HTTP_METHODS = HttpMethods

    def __init__(self, **kwargs):
        self._kwargs = kwargs
        self.client: Client | AsyncClient | None = None
        self.cookies: Cookies | None = None
        self.auth_flow: AuthFlow | None = None
        self.headers: dict = {}
        self.auth_params: dict = {}
        self.is_logged_in: bool = False
        self.user_id: int | None = None
        self.contact_id: int | None = None
        self.limits: Limits | None = None
        self._client_class: type(Client) | type(AsyncClient) | None = None
        self._api_key: str | None = None
        self._site_key: str | None = None
        self._username: str | None = None
        self._password: str | None = None
        self._base_url: str | None = None
        self._url_api4: str | None = None
        self._url_api3: str | None = None
        self._url_auth_login: str | None = None
        self._basic_auth: tuple | None = None
        self._timeout: int = 300
        self._ignore_ssl: bool = False
        self._call_count: int = 0
        self._is_ready: bool = False

    def setup(
        self,
        url: str,
        auth_flow: AuthFlow,
        api_key: str,
        **kwargs,
    ):
        """
        Basic setup for the client.
        :param url:
        :param auth_flow:
        :param api_key:
        :param kwargs:
        :return:
        """

        # Check URL
        if not validators.url(url):
            raise IOError(f"'{url}' is not a valid URL")

        # Check SSL
        self._ignore_ssl = kwargs.get("ignore_ssl", False)
        if not url.startswith("https"):
            logger.warning("The provided URL does not start with 'https'!")
            if not self._ignore_ssl:
                raise SslException()

        # Set up url and api endpoints
        self._base_url = url
        self._url_api3 = "/civicrm/ajax/rest"
        self._url_api4 = "/civicrm/ajax/api4"
        self._url_auth_login = "/civicrm/authx/login"

        # Gather authentication details
        self.auth_flow = auth_flow
        self._api_key = api_key
        self._site_key = kwargs.get("site_key", None)
        self._username = kwargs.get("username", None)
        self._password = kwargs.get("password", None)

        # Set up authentication
        self._set_up_authentication()

        # Set up basic auth
        self._basic_auth = (
            BasicAuth(*kwargs.get("basic_auth"))
            if kwargs.get("basic_auth", False)
            else None
        )

        # Configure timeout
        self._timeout = kwargs.get("timeout", 300)

        # Configure number of concurrent connections
        self.limits = Limits(max_connections=kwargs.get("max_connections", 10))

        # Configure cookies
        if not self.cookies:
            self.cookies = Cookies()

        # Create client
        self.client = self._client_class(
            base_url=self._base_url,
            headers=self.headers,
            verify=not self._ignore_ssl,
            cookies=self.cookies,
            auth=self._basic_auth,
            timeout=self._timeout,
            trust_env=kwargs.get("trust_env", False),
            follow_redirects=kwargs.get("follow_redirects", False),
            limits=self.limits,
        )

        self._is_ready = True

    def _set_up_authentication(self):
        """
        Set up authentication based on the selected auth flow.
        :return:
        """
        match self.auth_flow:
            case AuthFlow.XHEADER:
                self.headers["X-Civi-Auth"] = f"Bearer {self._api_key}"
                if self._site_key:
                    self.headers["X-Civi-Key"] = self._site_key
            case AuthFlow.COMMON:
                self.headers["Authorization"] = f"Bearer {self._api_key}"
                if self._site_key:
                    self.headers["X-Civi-Key"] = self._site_key
            case AuthFlow.PARAMETER:
                self.auth_params["_authx"] = f"Bearer+{self._api_key}"
                if self._site_key:
                    self.auth_params["_authxSiteKey"] = self._site_key
            case AuthFlow.LEGACY:
                if not self._site_key:
                    message = (
                        "Please provide a site key if you are using the "
                        "'legacy' authentication flow."
                    )
                    raise ApiAuthError(message)
                self.auth_params["api_key"] = self._api_key
                self.auth_params["key"] = self._site_key
            case AuthFlow.LOGIN_SESSION:
                if not self._username or not self._password:
                    message = (
                        "Please provide a username and a password if "
                        "you are using the 'login_session' "
                        "authentication flow"
                    )
                    raise ApiAuthError(message)
                user_password = f"{self._username}:{self._password}"
                cred_bytes = user_password.encode("utf-8")
                cred_b64 = b64encode(cred_bytes).decode("utf-8")
                self.auth_params["_authx"] = f"Basic+{cred_b64}"
                if self._site_key:
                    self.auth_params["_authxSiteKey"] = self._site_key
            case _:
                message = f"Authentication flow '{self.auth_flow}' is not " "supported."
                raise ApiAuthError(message)

    def _prepare_request(
        self,
        method: HttpMethods,
        url: str,
        params: dict = None,
        data: dict = None,
        json_data: dict = None,
    ):
        """
        Prepare the request.
        :return:
        """
        # Check if the client is ready
        if not self.is_ready():
            raise ApiException(
                "Client is not yet set up." "Please call 'connect()' first."
            )

        # Log the call for debugging purposes
        logger.debug(
            f"Preparing request to: {url}",
            extra={
                "method": method.value,
                "params": params,
                "data": data,
                "auth_flow": self.auth_flow,
                "timeout": self.client.timeout,
                "ignore_ssl": self._ignore_ssl,
                "call_count": self._call_count,
            },
        )

        # Build the request
        return self.client.build_request(
            method.value,
            url,
            headers=self.headers,
            params=self.auth_params | (params or {}),
            data=data,
            json=json_data,
        )

    @abstractmethod
    def _send_request(self, request):
        """
        Send the request.
        Must be implemented by the child class synchronously or asynchronously.
        :return:
        """
        pass

    @staticmethod
    def _handle_response(response):
        """
        Allgemeine Logik zur Behandlung von Antworten
        """
        response.raise_for_status()
        content = response.json()
        if content.get("is_error"):
            raise ApiException(content.get("error_message"))
        return content

    def api3(
        self,
        entity: str,
        action: str,
        query: dict,
        method: HttpMethods = None,
        **kwargs,
    ):
        """
        Execute an APIv3 call.
        :param entity:
        :param action:
        :param query:
        :param method: The HTTP method to use
        :return:
        :return:
        """
        url = self._url_api3
        data = None
        params = {
            "entity": entity,
            "action": action,
            "sequential": 1,
            "json": 1,
        }
        if method is None:
            method = get_method(action)

        # In case of a GET request, data has to be submitted as query params
        if method == self.HTTP_METHODS.GET and query:
            params.update(query)
        # In case of a POST request, data can be submitted form-encoded
        else:
            data = query

        request = self._prepare_request(method, url, params=params, data=data)
        result = self._send_request(request)
        return self._handle_response(result)

    def execute_api4(self, method: HttpMethods, entity: Api4):
        """
        Execute an APIv4 call by handing an APIv4 instance.
        Better call do not call this method directly but via Entity.execute()
        :param method:
        :param entity: An instance of Api4
        :return:
        """
        url = self._url_api4 + f"/{entity.entity}/{entity.action}"
        data = entity.build_query()  # Todo: Does not work yet

        request = self._prepare_request(method, url, json_data=data)
        result = self._send_request(request)
        return self._handle_response(result)

    def is_ready(self):
        """
        Check if the client is ready.
        :return:
        """
        return self._is_ready

    @abstractmethod
    def disconnect(self):
        """
        Disconnect from the API.
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def login(self):
        """
        Log in if the auth flow is login session.
        :return:
        """
        raise NotImplementedError


class CiviCrmApi(BaseCiviCrmApi):
    """
    This class provides an interface to CiviCRM APIv4.
    In most cases, you should not instantiate this class directly but use the
    global api instance from the package.
    """

    def setup(
        self,
        url: str,
        auth_flow: AuthFlow,
        api_key: str,
        **kwargs,
    ) -> None:

        self._client_class = Client

        # Base setup
        super().setup(url, auth_flow, api_key, **kwargs)

    def disconnect(self):
        """
        Close the client and clear the cookies.
        :return:
        """
        self.client.close()
        self.cookies.clear()
        self.is_logged_in = False
        logger.info("Client closed")

    def login(self):

        if not self.is_ready():
            raise ApiException(
                "Client is not yet set up." "Please call 'connect()' first."
            )

        # Log the call for debugging purposes
        logger.debug(
            f"API call to: {self._url_auth_login}",
            extra={
                "method": "get",
                "auth_flow": self.auth_flow,
                "timeout": self.client.timeout,
                "ignore_ssl": self._ignore_ssl,
                "call_count": self._call_count,
            },
        )

        # Make the login call
        try:
            response = self.client.get(self._url_auth_login)
            response.raise_for_status()
            content = response.json()
            if content.get("flow", False) == "login":
                if "user_id" in content.keys():
                    self.user_id = content["user_id"]
                if "contact_id" in content.keys():
                    self.contact_id = content["contact_id"]
                logger.info("Login successful.")
                self.is_logged_in = True
            else:
                m = "Authentication failed."
                logger.exception(m, exc_info=True)
                raise ApiAuthError(m)

        # Catch errors
        except Exception as e:
            m = f"Login failed: {e}"
            logger.exception(m, exc_info=True)
            raise ApiAuthError(str(e))

        finally:
            self._call_count += 1

    def _send_request(self, request: Request):
        """
        Synchronous implementation of sending a request
        :param request:
        :return: Response
        """
        try:
            result = self.client.send(request)

        # Catch exceptions
        except Exception as e:
            logger.exception(e)
            raise ApiError(str(e))

        # Increment call count
        finally:
            self._call_count += 1

        return result


class AsyncCiviCrmApi(BaseCiviCrmApi):
    """
    This class provides an async interface to CiviCRM APIv4.
    In most cases, you should not instantiate this class directly but use the
    global async_api instance from the package.
    """

    def setup(
        self,
        url: str,
        auth_flow: AuthFlow,
        api_key: str,
        **kwargs,
    ):
        self._client_class = AsyncClient

        # Base setup
        super().setup(url, auth_flow, api_key, **kwargs)

    async def disconnect(self):
        """
        Close the client and clear the cookies.
        :return:
        """
        await self.client.aclose()
        self.cookies.clear()
        self.is_logged_in = False
        logger.info("Client closed")

    async def login(self):

        if not self.is_ready():
            raise ApiException(
                "Client is not yet set up." "Please call 'connect()' first."
            )

        # Log the call for debugging purposes
        logger.debug(
            f"API call to: {self._url_auth_login}",
            extra={
                "method": "get",
                "auth_flow": self.auth_flow,
                "timeout": self.client.timeout,
                "ignore_ssl": self._ignore_ssl,
                "call_count": self._call_count,
            },
        )

        # Make the login call
        try:
            response = await self.client.get(self._url_auth_login)
            response.raise_for_status()
            content = await response.json()
            if content.get("flow", False) == "login":
                if "user_id" in content.keys():
                    self.user_id = content["user_id"]
                if "contact_id" in content.keys():
                    self.contact_id = content["contact_id"]
                logger.info("Login successful.")
                self.is_logged_in = True
            else:
                m = "Authentication failed."
                logger.exception(m)
                raise ApiAuthError(m)

        # Catch errors
        except Exception as e:
            m = f"Login failed: {e}"
            logger.exception(m)
            raise ApiAuthError(str(e))

        finally:
            self._call_count += 1

    async def _send_request(self, request: Request):
        """
        Asynchronous implementation of sending a request
        :param request:
        :return: Response
        """
        try:
            result = await self.client.send(request)

        # Catch exceptions
        except Exception as e:
            logger.exception(e, exc_info=True)
            raise ApiError(str(e))

        # Increment call count
        finally:
            self._call_count += 1

        return result
